@@ -5,17 +5,18 @@
 #include "commit.h"
 #include "fetch.h"
 
-static int use_link = 0;
-static int use_symlink = 0;
+static int use_link;
+static int use_symlink;
 static int use_filecopy = 1;
+static int commits_on_stdin;
 
-static char *path; /* "Remote" git repository */
+static const char *path; /* "Remote" git repository */
 
 void prefetch(unsigned char *sha1)
 {
 }
 
-static struct packed_git *packs = NULL;
+static struct packed_git *packs;
 
 static void setup_index(unsigned char *sha1)
 {
@@ -42,8 +43,8 @@ static int setup_indices(void)
 		return -1;
 	while ((de = readdir(dir)) != NULL) {
 		int namelen = strlen(de->d_name);
-		if (namelen != 50 || 
-		    strcmp(de->d_name + namelen - 5, ".pack"))
+		if (namelen != 50 ||
+		    !has_extension(de->d_name, ".pack"))
 			continue;
 		get_sha1_hex(de->d_name + 5, sha1);
 		setup_index(sha1);
@@ -63,9 +64,9 @@ static int copy_file(const char *source, char *dest, const char *hex,
 		}
 		/* If we got ENOENT there is no point continuing. */
 		if (errno == ENOENT) {
-			if (warn_if_not_exists)
-				fprintf(stderr, "does not exist %s\n", source);
-			return -1;
+			if (!warn_if_not_exists)
+				return -1;
+			return error("does not exist %s", source);
 		}
 	}
 	if (use_symlink) {
@@ -73,9 +74,8 @@ static int copy_file(const char *source, char *dest, const char *hex,
 		if (stat(source, &st)) {
 			if (!warn_if_not_exists && errno == ENOENT)
 				return -1;
-			fprintf(stderr, "cannot stat %s: %s\n", source,
-				strerror(errno));
-			return -1;
+			return error("cannot stat %s: %s", source,
+				     strerror(errno));
 		}
 		if (!symlink(source, dest)) {
 			pull_say("symlink %s\n", hex);
@@ -89,25 +89,21 @@ static int copy_file(const char *source, char *dest, const char *hex,
 		if (ifd < 0) {
 			if (!warn_if_not_exists && errno == ENOENT)
 				return -1;
-			fprintf(stderr, "cannot open %s\n", source);
-			return -1;
+			return error("cannot open %s", source);
 		}
 		ofd = open(dest, O_WRONLY | O_CREAT | O_EXCL, 0666);
 		if (ofd < 0) {
-			fprintf(stderr, "cannot open %s\n", dest);
 			close(ifd);
-			return -1;
+			return error("cannot open %s", dest);
 		}
 		status = copy_fd(ifd, ofd);
 		close(ofd);
 		if (status)
-			fprintf(stderr, "cannot write %s\n", dest);
-		else
-			pull_say("copy %s\n", hex);
-		return status;
+			return error("cannot write %s", dest);
+		pull_say("copy %s\n", hex);
+		return 0;
 	}
-	fprintf(stderr, "failed to copy %s with given copy methods.\n", hex);
-	return -1;
+	return error("failed to copy %s with given copy methods.", hex);
 }
 
 static int fetch_pack(const unsigned char *sha1)
@@ -180,13 +176,11 @@ int fetch_ref(char *ref, unsigned char *sha1)
 	ifd = open(filename, O_RDONLY);
 	if (ifd < 0) {
 		close(ifd);
-		fprintf(stderr, "cannot open %s\n", filename);
-		return -1;
+		return error("cannot open %s", filename);
 	}
-	if (read(ifd, hex, 40) != 40 || get_sha1_hex(hex, sha1)) {
+	if (read_in_full(ifd, hex, 40) != 40 || get_sha1_hex(hex, sha1)) {
 		close(ifd);
-		fprintf(stderr, "cannot read from %s\n", filename);
-		return -1;
+		return error("cannot read from %s", filename);
 	}
 	close(ifd);
 	pull_say("ref %s\n", sha1_to_hex(sha1));
@@ -194,17 +188,19 @@ int fetch_ref(char *ref, unsigned char *sha1)
 }
 
 static const char local_pull_usage[] =
-"git-local-fetch [-c] [-t] [-a] [-d] [-v] [-w filename] [--recover] [-l] [-s] [-n] commit-id path";
+"git-local-fetch [-c] [-t] [-a] [-v] [-w filename] [--recover] [-l] [-s] [-n] [--stdin] commit-id path";
 
-/* 
+/*
  * By default we only use file copy.
  * If -l is specified, a hard link is attempted.
  * If -s is specified, then a symlink is attempted.
  * If -n is _not_ specified, then a regular file-to-file copy is done.
  */
-int main(int argc, char **argv)
+int main(int argc, const char **argv)
 {
-	char *commit_id;
+	int commits;
+	const char **write_ref = NULL;
+	char **commit_id;
 	int arg = 1;
 
 	setup_git_directory();
@@ -229,21 +225,30 @@ int main(int argc, char **argv)
 		else if (argv[arg][1] == 'v')
 			get_verbosely = 1;
 		else if (argv[arg][1] == 'w')
-			write_ref = argv[++arg];
+			write_ref = &argv[++arg];
 		else if (!strcmp(argv[arg], "--recover"))
 			get_recover = 1;
+		else if (!strcmp(argv[arg], "--stdin"))
+			commits_on_stdin = 1;
 		else
 			usage(local_pull_usage);
 		arg++;
 	}
-	if (argc < arg + 2)
+	if (argc < arg + 2 - commits_on_stdin)
 		usage(local_pull_usage);
-	commit_id = argv[arg];
-	path = argv[arg + 1];
-	write_ref_log_details = path;
+	if (commits_on_stdin) {
+		commits = pull_targets_stdin(&commit_id, &write_ref);
+	} else {
+		commit_id = (char **) &argv[arg++];
+		commits = 1;
+	}
+	path = argv[arg];
 
-	if (pull(commit_id))
+	if (pull(commits, commit_id, write_ref, path))
 		return 1;
 
+	if (commits_on_stdin)
+		pull_targets_free(commits, commit_id, write_ref);
+
 	return 0;
 }
